/*
 * Copy me if you can.
 * by 20h
 */

#include <u.h>
#include <libc.h>
#include <draw.h>
#include <event.h>
#include <stdio.h>

char *buttons[] = 
{
	"dot",
	"line",
	"rectangle",
	"ellipse",
	"write",
	"clear",
	"exit",
	0
};

enum {
	DOT = 0x0,
	LINE,
	RECTANGLE,
	ELLIPSE,
	WRITE,
	CLEAR,
	EXIT
};

Menu menu = 
{
	buttons
};

Image *b, *w;

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, b, nil, ZP);

	return;
}

void
usage(void)
{
	print("usage: 9paint [-re] [-t thick] [file]\n");
	return;
}

int
dist_points(Point a, Point b)
{

	return (int)sqrt((a.x - b.x) * (a.x - b.x) + (a.y - b.y) * (a.y - b.y));
}

/* Taken from /ys/src/cmd/page/view.c */
int 
max(int a, int b)
{
	return a > b ? a : b;
}

Rectangle
screenrect(void)
{
	int fd;
	char buf[12*5];

	fd = open("/dev/screen", OREAD);
	if(fd == -1)
		fd=open("/mnt/term/dev/screen", OREAD);
	if(fd == -1){
		fprint(2, "page: can't open /dev/screen: %r\n");
		exits("window read");
	}
	if(read(fd, buf, sizeof buf) != sizeof buf){
		fprint(2, "page: can't read /dev/screen: %r\n");
		exits("screen read");
	}
	close(fd);
	return Rect(atoi(buf+12), atoi(buf+24), atoi(buf+36), atoi(buf+48));
}

/* Taken from /sys/src/cmd/page/view.c */
Image *
nullize_img(Image *im, int o)
{
	Image *tmp;

	if(im->r.min.x > 0 || im->r.min.y > 0) {
		tmp = allocimage(display, Rect(0, 0, Dx(im->r), Dy(im->r)), im->chan, 0, DNofill);
		if(tmp == nil) {
			fprint(2, "xallocimage: %r\n");
			exits("memory");
		}
		drawop(tmp, tmp->r, im, nil, im->r.min, S);
		if(o == 0)
			freeimage(im);
		im = tmp;
	}

	return im;
}

void
resize(int dx, int dy)
{
	static Rectangle sr;
	Rectangle r, or;
	int wctlfd;

	dx += 2*Borderwidth;
	dy += 2*Borderwidth;

	wctlfd = open("/dev/wctl", OWRITE);
	if(wctlfd < 0)
		return;

	r = insetrect(screen->r, -Borderwidth);
	if(Dx(r) == dx && Dy(r) == dy)
	{
		close(wctlfd);
		return;
	}

	if(Dx(sr)*Dy(sr) == 0)
		sr = screenrect();

	or = r;

	r.max.x = r.min.x+dx;
	r.max.y = r.min.y+dy;
	if(r.max.x > sr.max.x){
		if(Dx(r) > Dx(sr)){
			r.min.x = 0;
			r.max.x = sr.max.x;
		}else{
			r = rectaddpt(r, Pt(sr.max.x-r.max.x, 0));
		}
	}
	if(r.max.y > sr.max.y){
		if(Dy(r) > Dy(sr)){
			r.min.y = 0;
			r.max.y = sr.max.y;
		}else{
			r = rectaddpt(r, Pt(0, sr.max.y-r.max.y));
		}
	}

	if(Dx(r) == Dx(or) && Dy(r) == Dy(or))
	{
		close(wctlfd);
		return;
	}


	fprint(wctlfd, "resize -minx %d -miny %d -maxx %d -maxy %d\n",
		r.min.x, r.min.y, r.max.x, r.max.y);
	close(wctlfd);
}

void 
main(int argc, char **argv)
{
	Mouse m;
	Point l;
	char mode, set, re;
	int fd, t;
	Image *e;

	mode = 0;
	l.x = -1;
	l.y = -1;
	t = 0;
	re = 0;

	if(initdraw(nil, nil, "9paint") < 0)
		sysfatal("initdraw failed: %r");

	b = display->black;
	w = display->white;

	ARGBEGIN {
		case 'r':
			b = display->white;
			w = display->black;
			break;
		case 't':
			t = atoi(ARGF());
			break;
		case 'e':
			re = 1;
			break;
		default:
			usage();
			exits(0);
			break;
	} ARGEND;

	eresized(0);

	if(*argv != nil)
	{
		fd = open(*argv, OREAD);
		if(fd < 0)
		{
			perror("open");
		} else {
			e = readimage(display, fd, 0);
			if(e != nil)
			{
				e = nullize_img(e, 0);
				if(re == 1)
					resize(Dx(e->r), Dy(e->r));
				eresized(1);
				drawop(screen, screen->r, e, nil, e->r.min, S);
			} else {
				perror("readimage");
			}

			close(fd);
		}
	}

	einit(Emouse);

	while(1) {
		if(ecanmouse()) {
			m = emouse();
			set = 0;

			if(m.buttons & 4)
			{
				switch((mode = emenuhit(3, &m, &menu)))
				{
					case 0:
					case 1:
					case 2:
					case 3:
						l.x = -1;
						l.y = -1;
						break;
					case WRITE:
						fd = open("9paint.img", OWRITE);
						if(fd < 0)
						{
							fd = create("9paint.img", OWRITE, 0777);
							if(fd < 0)
							{
								perror("create");
								break;
							}
						}
						
						e = nullize_img(screen, 2);
						if(writeimage(fd, e, 0) < 0)
						{
							perror("writeimage");
							break;
						}
						freeimage(e);

						close(fd);
						break;
					case CLEAR:
						draw(screen, screen->r, b, nil, ZP);
						l.x = -1;
						l.y = -1;
						break;
					case EXIT:
						exits(0);
						break;
					default:
						break;
				}
				continue;
			}
			if(m.buttons & 1)
				set = 1;
			if(m.buttons & 2)
				set = 2;
			if(set > 0)
			{
				if(l.x == -1 && l.y == -1 && mode != DOT)
				{
					l = m.xy;
					continue;
				}

				switch(mode)
				{
					case DOT:
						fillellipse(screen, m.xy, t, t, 
							(set == 1) ? w : b, ZP);
						break;
					case LINE:
						line(screen, l, m.xy, 0, 0, 1,
							(set == 1) ? w : b, ZP);
						l = m.xy;
						break;
					case RECTANGLE:
						line(screen, l, (Point){m.xy.x, l.y}, Endsquare, Endsquare, t,
							(set == 1) ? w : b, ZP);
						line(screen, l, (Point){l.x, m.xy.y}, Endsquare, Endsquare, t,
							(set == 1) ? w : b, ZP);
						line(screen, m.xy, (Point){m.xy.x, l.y}, Endsquare, Endsquare, t,
							(set == 1) ? w : b, ZP);
						line(screen, m.xy, (Point){l.x, m.xy.y}, Endsquare, Endsquare, t,
							(set == 1) ? w : b, ZP);
						l = m.xy;
						break;
					case ELLIPSE:
						ellipse(screen, l, dist_points(l, m.xy), dist_points(l, m.xy), t, 
							(set == 1) ? w : b, ZP);
						l = m.xy;
						break;
					default:
						break;
				}
			}
		}
		sleep(1);
	}

	exits(0);
}